/* 
 * File:   lc_meter.c
 * Author: k15
 *
 * Created on 11. prosinec 2013, 16:20
 */

// -----------------------------------------------------------------------
//	LCR meter
//	K15 2011
// -----------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>

/*
 * 
 */

// calibration values:

const float low_current	= 98.5;		// low current source [uA]
const float high_current = 91600.0;	// high current source[uA]

// A/D converter resolutions [uV]:
const float low_ad_res	= 51.1056511056511;
const float mid_ad_res  = 500.616522811344;
const float high_ad_res	= 5033.93665158371;

// counter interval [us]:
const float counter_interval = 6.4;
// ESR overcharge time [us]:
const float esr_charge_time = 0.1;

// serial induktance in LC oscillator [uH]:
float lc_Ls = 10.0;

const float lc_Q = 2676852.091;

// parallel capacitance in RC oscillator [pF]:
float rc_Cp = 1000.0;

const float rc_Q = 318309.88614;

// serial resistance in ohmmeter:
float r_Rs = 0;

// serial resistance in ESR meter:
float e_Rs = 0;

// EEPROM addresses of "zero" quocients
#define Ls_addr	0x00
#define Cp_addr	0x04
#define Rs_addr	0x08
#define Es_addr	0x0C

/* ----------------------------------------------------------------------- */
#include <pic18f252.h>
#include <stdio.h>
#include <lcd.h>
#include <delay.h>
#include <word.h>
#include <hex.h>
#include <eeprom.h>

/* ----------------------------------------------------------------------- */

/* Konfigurační bity: upravte dle vašeho nastavení a potřeb */
/* PIKLAB:
__code char __at __CONFIG1H CONFIG1H = _OSC_HS_PLL_1H & _OSCS_OFF_1H;
__code char __at __CONFIG2L CONFIG2L = _BODEN_OFF_2L;
__code char __at __CONFIG2H CONFIG2H = _WDT_OFF_2H;
__code char __at __CONFIG3H CONFIG3H = _CCP2MUX_RC1_3H;
__code char __at __CONFIG4L CONFIG4L = _STVR_ON_4L & _LVP_OFF_4L & _BACKBUG_OFF_4L;
__code char __at __CONFIG5L CONFIG5L = _CP_0_OFF_5L & _CP_1_OFF_5L & _CP_2_OFF_5L & _CP_3_OFF_5L;
__code char __at __CONFIG5H CONFIG5H = _CPB_OFF_5H & _CPD_OFF_5H;
__code char __at __CONFIG6L CONFIG6L = _WRT_0_OFF_6L & _WRT_1_OFF_6L & _WRT_2_OFF_6L & _WRT_3_OFF_6L;
__code char __at __CONFIG6H CONFIG6H = _WRTC_OFF_6H & _WRTB_OFF_6H & _WRTD_OFF_6H;
__code char __at __CONFIG7L CONFIG7L = _EBTR_0_OFF_7L & _EBTR_1_OFF_7L & _EBTR_2_OFF_7L & _EBTR_3_OFF_7L;
__code char __at __CONFIG7H CONFIG7H = _EBTRB_OFF_7H;
*/

// PIC18F252 Configuration Bit Settings

#include <xc.h>

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

// CONFIG1H
#pragma config OSC = HSPLL      // Oscillator Selection bits (HS oscillator with PLL enabled/Clock frequency = (4 x FOSC))
#pragma config OSCS = OFF       // Oscillator System Clock Switch Enable bit (Oscillator system clock switch option is disabled (main oscillator is source))

// CONFIG2L
#pragma config PWRT = OFF       // Power-up Timer Enable bit (PWRT disabled)
#pragma config BOR = ON         // Brown-out Reset Enable bit (Brown-out Reset enabled)
#pragma config BORV = 20        // Brown-out Reset Voltage bits (VBOR set to 2.0V)

// CONFIG2H
#pragma config WDT = OFF        // Watchdog Timer Enable bit (WDT disabled (control is placed on the SWDTEN bit))
#pragma config WDTPS = 128      // Watchdog Timer Postscale Select bits (1:128)

// CONFIG3H
#pragma config CCP2MUX = ON     // CCP2 Mux bit (CCP2 input/output is multiplexed with RC1)

// CONFIG4L
#pragma config STVR = ON        // Stack Full/Underflow Reset Enable bit (Stack Full/Underflow will cause RESET)
#pragma config LVP = OFF        // Low Voltage ICSP Enable bit (Low Voltage ICSP disabled)

// CONFIG5L
#pragma config CP0 = OFF        // Code Protection bit (Block 0 (000200-001FFFh) not code protected)
#pragma config CP1 = OFF        // Code Protection bit (Block 1 (002000-003FFFh) not code protected)
#pragma config CP2 = OFF        // Code Protection bit (Block 2 (004000-005FFFh) not code protected)
#pragma config CP3 = OFF        // Code Protection bit (Block 3 (006000-007FFFh) not code protected)

// CONFIG5H
#pragma config CPB = OFF        // Boot Block Code Protection bit (Boot Block (000000-0001FFh) not code protected)
#pragma config CPD = OFF        // Data EEPROM Code Protection bit (Data EEPROM not code protected)

// CONFIG6L
#pragma config WRT0 = OFF       // Write Protection bit (Block 0 (000200-001FFFh) not write protected)
#pragma config WRT1 = OFF       // Write Protection bit (Block 1 (002000-003FFFh) not write protected)
#pragma config WRT2 = OFF       // Write Protection bit (Block 2 (004000-005FFFh) not write protected)
#pragma config WRT3 = OFF       // Write Protection bit (Block 3 (006000-007FFFh) not write protected)

// CONFIG6H
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers (300000-3000FFh) not write protected)
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot Block (000000-0001FFh) not write protected)
#pragma config WRTD = OFF       // Data EEPROM Write Protection bit (Data EEPROM not write protected)

// CONFIG7L
#pragma config EBTR0 = OFF      // Table Read Protection bit (Block 0 (000200-001FFFh) not protected from Table Reads executed in other blocks)
#pragma config EBTR1 = OFF      // Table Read Protection bit (Block 1 (002000-003FFFh) not protected from Table Reads executed in other blocks)
#pragma config EBTR2 = OFF      // Table Read Protection bit (Block 2 (004000-005FFFh) not protected from Table Reads executed in other blocks)
#pragma config EBTR3 = OFF      // Table Read Protection bit (Block 3 (006000-007FFFh) not protected from Table Reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      // Boot Block Table Read Protection bit (Boot Block (000000-0001FFh) not protected from Table Reads executed in other blocks)




#define RE_C		LATCbits.LATC4
#define RE_L		LATCbits.LATC3

#define BTN		LATCbits.LATC1

//#define CS_HI		LATAbits.LATA3
//#define CS_LO		LATAbits.LATA4
//#define DISCH		LATAbits.LATA5

// nastaveni registru LATA - mereni C+ESR - ovladani zdroju proudu:
#define C_ESR_HI_CURRENT	0x10
#define C_ESR_LO_CURRENT	0x08
#define C_ESR_OFF		0x18
#define C_ESR_DISCHARGE		0x38
#define C_ESR_DISCH_TEST	0x28

//#define CLRWDT		__asm clrwdt __endasm

// nastaveni registru ADCON0:
// C_ESR_AD_HI 	- 0x0318 = 4 V, tedy rozliseni AD prevodniku 0,50505 mV
// C_ESR_AD_MI 	- 0x0318 = 400 mV, tedy rozliseni AD prevodniku 0,50505 mV
// C_ESR_AD_LO 	- 0x0310 = 40 mV,  tedy rozliseni AD prevodniku 0,05102 mV
#define C_ESR_AD_LO		0x89
#define C_ESR_AD_MI		0x91
#define C_ESR_AD_HI		0x81

typedef enum {
	modeL,
	modeC,
	modeE,
	modeR,
	modeL_zero,
	modeC_zero,
	modeE_zero,
	modeR_zero,
	// self-diagnostics modes:
	modeTestCS_L,
	modeTestCS_H,
	modeTestDISCH
} operMode;

volatile operMode mode = modeC;

// result in kHz:
float measure_f () {
	K15WORD c_mes;
	K15WORD t_mes;

	T1CON = 0x86;	// 16bit, no prescaler, asynchronous external input
	TMR1H = 0;
	TMR1L = 0;
	PIR1bits.TMR1IF = 0;

	//T3CON = 0x80;	// 16bit, no prescaler, internal oscillator
	T3CON = 0xB0;	// 16bit, prescaler 1:8, internal oscillator
	TMR3H = 0;
	TMR3L = 0;
	PIR2bits.TMR3IF = 0;

	T1CONbits.TMR1ON = 1;
	T3CONbits.TMR3ON = 1;

	while (!PIR2bits.TMR3IF);

	T1CONbits.TMR1ON = 0;
	T3CONbits.TMR3ON = 0;

	c_mes.lo = TMR1L;
	c_mes.hi = TMR1H;
	if (c_mes.hi || !c_mes.word) {
		// the frequency is high enough, let's use the pulse count:
		return (float)c_mes.word / 52.4288;  // [10MHz]/8
	}

	// the frequency is low, let's measure duration of one period:
	T3CON = 0xF0;	// 16bit, prescaler 1:8, internal oscillator, source for CPP modules
	TMR3H = 0;
	TMR3L = 0;
	PIR2bits.TMR3IF = 0;
	PIR1bits.CCP1IF = 0;
	T3CONbits.TMR3ON = 1;

	while (!PIR1bits.CCP1IF && !PIR2bits.TMR3IF);

	//TMR3H = 0;
	TMR3L = 0;

	if (PIR2bits.TMR3IF) return 0;

	PIR1bits.CCP1IF = 0;
	while (!PIR1bits.CCP1IF && !PIR2bits.TMR3IF);

	if (PIR2bits.TMR3IF) return 0;

	t_mes.lo = CCPR1L;
	t_mes.hi = CCPR1H;

	return 1250.0 / (float)t_mes.word;  // [10MHz]/8
}

void run_l () {
	float freq, L;
	char str[24];

	RE_C = 0;
	RE_L = 1;

	freq = measure_f();
	if (freq)
		L = (lc_Q / freq / freq) - lc_Ls;
	else
		L = 0;	// just to shut up the complaining compiler...

	if (freq == 0)
		sprintf(str, "  L: ---        ");
	else if (L < 1000.0)
		sprintf(str, "  L: %6.3f uH   ", L);
	else if (L < 1000000.0)
		sprintf(str, "  L: %6.3f mH   ", L/1000.0);
	else 	sprintf(str, "  L: %6.3f  H   ", L/1000000.0);

	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string(str);

	sprintf(str, "(f: %6.3f kHz) ", freq);

	write_lcd_cmd(0xC0); 	// line 2
	print_lcd_string(str);
}

void run_l_zero () {
	float freq;

	RE_C = 0;
	RE_L = 1;

	freq = measure_f();

	if (freq) {
		lc_Ls = (lc_Q / freq / freq);
		eeprom_write_float(Ls_addr, &lc_Ls);
	}

	mode = modeL;
}

void run_c () {
	float freq, C;
	char str[24];

	RE_C = 1;
	RE_L = 0;

	freq = measure_f();

	if (freq)
		C = (rc_Q / freq) - rc_Cp;
	else
		C = 0;	// just to shut up the complaining compiler...

	if (freq == 0)
		sprintf(str, "  C: ---        ");
	else if (C < 1000.0)
		sprintf(str, "  C: %6.3f pF    ", C);
	else if (C < 1000000.0)
		sprintf(str, "  C: %6.3f nF    ", C/1000.0);
	else
		sprintf(str, "  C: %6.3f uF    ", C/1000000.0);

	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string(str);

	sprintf(str, "(f: %6.3f kHz) ", freq);

	write_lcd_cmd(0xC0); 	// line 2
	print_lcd_string(str);
}

void run_c_zero () {
	float freq;

	RE_C = 1;
	RE_L = 0;

	freq = measure_f();

	if (freq) {
		rc_Cp = (rc_Q / freq);
		eeprom_write_float(Cp_addr, &rc_Cp);
	}

	mode = modeC;
}

char discharge() {
	// discharge:
	LATA = C_ESR_DISCHARGE;

	T0CON = 0x85; // TIMER0 16bit, internal clock, prescaler 1:64
	TMR0H = 0;
	TMR0L = 0;
	INTCONbits.TMR0IF = 0;

	do {
		ADCON0bits.GO = 1;
		while (ADCON0bits.GO) ;
		if (INTCONbits.TMR0IF) {
			LATA = C_ESR_OFF;
			return 0;
		}
	} while (ADRESH != 0 || ADRESL != 0);
	LATA = C_ESR_OFF;
	return 1;
}

char precharge(unsigned char cssel, unsigned int minval) {
	K15WORD u_mes;
	// precharge:
	T0CON = 0x85; // TIMER0 16bit, internal clock, prescaler 1:64
	TMR0H = 0;
	TMR0L = 0;
	INTCONbits.TMR0IF = 0;

	LATA = cssel;

	do {
		ADCON0bits.GO = 1;
		while (ADCON0bits.GO) ;
		if (INTCONbits.TMR0IF) {
			LATA = C_ESR_OFF;
			return 0;
		}
		u_mes.lo = ADRESL;
		u_mes.hi = ADRESH;
	} while (u_mes.word < minval);

	if (ADRESH) {
		LATA = C_ESR_OFF;
		return 0;
	}
	else
		return 1;
}


float measure_c(unsigned char adsel, unsigned char cssel, float ad_resolution, float current) {
	K15WORD u_pre;
	K15WORD u_mes;
	K15WORD t_mes;

	ADCON0 = adsel;

	if (!discharge()) return -1.0;
	if (!precharge(cssel, 1)) return -2.0;

	u_pre.lo = ADRESL;
	u_pre.hi = ADRESH;

	t_mes.word = 0;
	TMR0H = 0;
	TMR0L = 0;
	INTCONbits.TMR0IF = 0;

	// charge and measure:
	while (!ADRESH && !(TMR0H & 0xC0)) {
		ADCON0bits.GO = 1;
		while (ADCON0bits.GO) ;
		//if ((ADRESH & 0x02) || INTCONbits.TMR0IF) break;
		u_mes.lo = ADRESL;
		u_mes.hi = ADRESH;
		t_mes.lo = TMR0L;
		t_mes.hi = TMR0H;
	}

	LATA = C_ESR_OFF;

	if (u_mes.word < 0x30) return -2.0;
	if (u_mes.word > 0x0320) return 0.0;

	u_mes.word -= u_pre.word;

	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	u_pre.lo = ADRESL;
	u_pre.hi = ADRESH;

	return current * (float)t_mes.word * counter_interval / (float)u_mes.word / ad_resolution;
}

float measure_cap() {
	float c;
	if ((c = measure_c(C_ESR_AD_HI, C_ESR_LO_CURRENT, high_ad_res, low_current))  > 0.0) return c;
	if ((c = measure_c(C_ESR_AD_MI, C_ESR_LO_CURRENT, mid_ad_res,  low_current))  > 0.0) return c;
	if ((c = measure_c(C_ESR_AD_LO, C_ESR_LO_CURRENT, low_ad_res,  low_current))  > 0.0) return c;
	if ((c = measure_c(C_ESR_AD_HI, C_ESR_HI_CURRENT, high_ad_res, high_current)) > 0.0) return c;
	if ((c = measure_c(C_ESR_AD_MI, C_ESR_HI_CURRENT, mid_ad_res,  high_current)) > 0.0) return c;
	     c = measure_c(C_ESR_AD_LO, C_ESR_HI_CURRENT, low_ad_res,  high_current); 	     return c;
}

float measure_e(unsigned char adsel, unsigned char cssel, float ad_resolution, float current) {
	K15WORD u_pre;
	K15WORD u_mes;

	ADCON0 = adsel;

	if (!discharge()) return -1.0;
	if (!precharge(cssel, 0x0030)) return -2.0;

	// charge and measure:
	ADCON0bits.GO = 1;
	LATA = C_ESR_OFF;
	while (ADCON0bits.GO) ;
	if (ADRESH == 0x03) return 0.0;
	u_mes.lo = ADRESL;
	u_mes.hi = ADRESH;

	// measure:
	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	if (ADRESH == 0x03) return 0.0;
	u_pre.lo = ADRESL;
	u_pre.hi = ADRESH;

	u_mes.word -= u_pre.word;

/*
	// leakage correction:
	// the idea is the capacitor discharges through the circuit leakages. We will do one more measurement
	// and eliminate the difference.
	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	if (ADRESH == 0x03) return 0.0;

	u_mes.word -= u_pre.word;

	u_pre.lo = ADRESL;
	u_pre.hi = ADRESH;

	u_mes.word += u_pre.word;
*/

	return 1000.0 * (float)u_mes.word * ad_resolution / current;
}

float measure_esr() {
	float e;
	if ((e = measure_e(C_ESR_AD_LO, C_ESR_HI_CURRENT, low_ad_res,  high_current)) > 0.0) return e;
	if ((e = measure_e(C_ESR_AD_MI, C_ESR_HI_CURRENT, mid_ad_res,  high_current)) > 0.0) return e;
	if ((e = measure_e(C_ESR_AD_HI, C_ESR_HI_CURRENT, high_ad_res, high_current)) > 0.0) return e;
	if ((e = measure_e(C_ESR_AD_LO, C_ESR_LO_CURRENT, low_ad_res,  low_current))  > 0.0) return e;
	if ((e = measure_e(C_ESR_AD_MI, C_ESR_LO_CURRENT, mid_ad_res,  low_current))  > 0.0) return e;
	     e = measure_e(C_ESR_AD_HI, C_ESR_LO_CURRENT, high_ad_res, low_current);	     return e;
}

void run_e () {
	char str[24];
	float c, e, e_corr;

	RE_C = 0;
	RE_L = 0;

	c = measure_cap();

	if (c == -1.0)
		sprintf(str, "DISCHARGE ERROR!");
	else if (c == -2.0)
		sprintf(str, " SHORT CIRCUIT! ");
	else if (c == 0.0)
		sprintf(str, "  C: ---        ");
	else if (c < 1000.0)
		sprintf(str, "  C: %6.3f uF  ", c);
	else if (c < 1000000.0)
		sprintf(str, "  C: %6.3f mF  ", c/1000.0);
	else if (c < 10000000.0)
		sprintf(str, "  C: %6.3f  F  ", c/1000000.0);
	else
		sprintf(str, "  C: ---        ");

	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string(str);

	e = measure_esr();
	e_corr = e - e_Rs;

	if (c < 1.0)
		sprintf(str, "ESR: ---        ");
	else if (e <= 0.0)
		sprintf(str, "ESR: ---        ");
	else if (e_corr < 1000.0)
		sprintf(str, "ESR: %6.3f m\xF4  ", e_corr);
	else if (e_corr < 1000000.0)
		sprintf(str, "ESR: %6.3f  \xF4  ", e_corr/1000.0);
	else
		sprintf(str, "ESR: ---        ");

	write_lcd_cmd(0xC0); 	// line 2
	print_lcd_string(str);
}


void run_e_zero () {
	// the same as R zero...
	float e;

	RE_C = 0;
	RE_L = 0;

	e = measure_esr();

	if (e > 0) {
		e_Rs = e;
	 	eeprom_write_float(Es_addr, &e);
	}

	mode = modeE;
}


float measure_r(unsigned char adsel, unsigned char cssel, float ad_resolution, float current) {
	K15WORD u_mes;

	ADCON0 = adsel;
	LATA = cssel;

	delay(20);

	// measure:
	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	u_mes.lo = ADRESL;
	u_mes.hi = ADRESH;

	LATA = C_ESR_OFF;

	if (u_mes.hi == 0x03) return -1;

	if (u_mes.word < 0x30) return 0;

	return 1000.0 * (float)u_mes.word * ad_resolution / current;
}

float measure_res() {
	float r;
	RE_C = 0;
	RE_L = 0;

	if ((r = measure_r(C_ESR_AD_HI, C_ESR_LO_CURRENT, high_ad_res, low_current))  > 0) return r;
	if ((r = measure_r(C_ESR_AD_MI, C_ESR_LO_CURRENT, mid_ad_res,  low_current))  > 0) return r;
	if ((r = measure_r(C_ESR_AD_LO, C_ESR_LO_CURRENT, low_ad_res,  low_current))  > 0) return r;
	if ((r = measure_r(C_ESR_AD_HI, C_ESR_HI_CURRENT, high_ad_res, high_current)) > 0) return r;
	if ((r = measure_r(C_ESR_AD_MI, C_ESR_HI_CURRENT, mid_ad_res,  high_current)) > 0) return r;
	     r = measure_r(C_ESR_AD_LO, C_ESR_HI_CURRENT, low_ad_res,  high_current);	   return r;
}

void run_r () {
	float r, r_corr;
	char str[24];

	r = measure_res();
	r_corr = r - r_Rs;

	if (r <= 0)
		sprintf(str, "  R:   ---      ");
	else if (r_corr < 1000.0)
		sprintf(str, "  R: %6.3f m\xF4  ", r_corr);
	else if (r_corr < 1000000.0)
		sprintf(str, "  R: %6.3f  \xF4  ", r_corr/1000.0);
	else if (r_corr < 10000000.0)
		sprintf(str, "  R: %6.3f k\xF4  ", r_corr/1000000.0);
	else
		sprintf(str, "  R:   > 10 k\xF4  ");

	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string(str);

	write_lcd_cmd(0xC0); 	// line 2
	print_lcd_string("                ");
}

void run_r_zero () {
	float r;

	RE_C = 0;
	RE_L = 0;

	r = measure_res();

	if (r > 0) {
		r_Rs = r;
	 	eeprom_write_float(Rs_addr, &r);
	}

	mode = modeR;
}

void test_AD() {

	write_lcd_cmd(0xC0); 	// line 2

	ADCON0 = C_ESR_AD_HI;
	delay(20);
	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	write_lcd_data(hex[ADRESH >> 4]);
	write_lcd_data(hex[ADRESH & 0x0F]);
	write_lcd_data(hex[ADRESL >> 4]);
	write_lcd_data(hex[ADRESL & 0x0F]);

	ADCON0 = C_ESR_AD_MI;
	delay(20);
	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	print_lcd_string(" ");
	write_lcd_data(hex[ADRESH >> 4]);
	write_lcd_data(hex[ADRESH & 0x0F]);
	write_lcd_data(hex[ADRESL >> 4]);
	write_lcd_data(hex[ADRESL & 0x0F]);

	ADCON0 = C_ESR_AD_LO;
	delay(20);
	ADCON0bits.GO = 1;
	while (ADCON0bits.GO) ;
	print_lcd_string(" ");
	write_lcd_data(hex[ADRESH >> 4]);
	write_lcd_data(hex[ADRESH & 0x0F]);
	write_lcd_data(hex[ADRESL >> 4]);
	write_lcd_data(hex[ADRESL & 0x0F]);

	print_lcd_string(" ");
}


void run_test_cs_l () {
	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string("Test 100uA      ");
	RE_C = 0;
	RE_L = 0;

	LATA = C_ESR_LO_CURRENT;
	// 133 uA

	test_AD();
}

void run_test_cs_h () {
	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string("Test 100mA      ");
	RE_C = 0;
	RE_L = 0;
	LATA = C_ESR_HI_CURRENT;
	// 124 mA

	test_AD();
}

void run_test_disch () {
	write_lcd_cmd(0x80); 	// line 1
	print_lcd_string("Discharge test  ");
	RE_C = 0;
	RE_L = 0;
	LATA = C_ESR_DISCH_TEST;

	test_AD();
}

void isr(void) __interrupt (1)
{
	if (PIR2bits.CCP2IF) {
		LATA = C_ESR_OFF;
		delay(0xFFFF);
		TMR0H = 0;
		TMR0L = 0;
		INTCONbits.TMR0IF = 0;
		T0CON = 0x85; // TIMER0 16bit, internal clock, prescaler 1:64
		while (!PORTCbits.RC1 && !INTCONbits.TMR0IF);
		//clear_lcd();
		if (!INTCONbits.TMR0IF) {
			if (mode == modeR)
				mode = modeL;
			else if	(mode == modeTestDISCH)
				mode = modeTestCS_L;
			else
				mode++;
		} else {
			switch (mode) {
				case modeC: 	mode = modeC_zero; break;
				case modeE: 	mode = modeE_zero; break;
				case modeL: 	mode = modeL_zero; break;
				case modeR: 	mode = modeR_zero; break;
			}
		}
		//if (mode == modeTestDISCH) mode = modeL; else mode++;
		PIR2bits.CCP2IF = 0;
	}
	INTCONbits.GIE = 1;
}

int main(int argc, char** argv) {

	LATA = C_ESR_OFF;
	TRISA = 0x07;		// port A - RA0, RA1, RA2 inputs, the rest outputs
	TRISB = 0x00;		// LCD display data output
	TRISC = 0x07;		// port C - RC0, RC1, RC2 inputs, the rest outputs
	CCP1CON = 0x04;		// CPP1 in capture mode, falling edge - used for low frequency measurement
	CCP2CON = 0x04;		// CPP2 in capture mode, falling edge - used for mode button

	ADCON1 = 0xC0;

	PIE2bits.CCP2IE = 1;	// enable interrupt from RC1 input (CCP2 capture)
	INTCONbits.GIE = 1;	// enable interrupts
	INTCONbits.PEIE = 1;	// enable peripheral interrupts

	eeprom_read_float(Cp_addr, &rc_Cp);
	eeprom_read_float(Ls_addr, &lc_Ls);
	eeprom_read_float(Rs_addr, &r_Rs);
	eeprom_read_float(Es_addr, &e_Rs);

	// when the button is presses while switching ON, go to the diagnostics menu:
	if (!PORTCbits.RC1)
		mode = modeTestDISCH;
	else
		mode = modeE;

	init_lcd();

	while (1) {
		switch (mode) {
			case modeL: 		run_l(); break;
			case modeC: 		run_c(); break;
			case modeE: 		run_e(); break;
			case modeR: 		run_r(); break;
			case modeTestCS_L: 	run_test_cs_l(); break;
			case modeTestCS_H: 	run_test_cs_h(); break;
			case modeTestDISCH: 	run_test_disch(); break;
			case modeL_zero: 	run_l_zero(); break;
			case modeC_zero: 	run_c_zero(); break;
			case modeE_zero: 	run_e_zero(); break;
			case modeR_zero: 	run_r_zero(); break;
			default:
					// this should never happen:
					write_lcd_cmd(0x80); 	// line 1
					print_lcd_string("* select mode **");
					write_lcd_cmd(0xC0); 	// line 2
					print_lcd_string("****************");
		}
		CLRWDT();
	}
            return (EXIT_SUCCESS);

}
