#include <pic18f252.h>
#include <stdio.h>
#include <delay.h>
#include <word.h>
#include <hex.h>


#define LCD_E 		LATCbits.LATC7
#define LCD_RW 		LATCbits.LATC6
#define LCD_RS 		LATCbits.LATC5

#define LCD_INPUT	TRISB=0xFF
#define LCD_OUTPUT	TRISB=0x00

#define LCD_OUT		LATB

#define LCD_BF		PORTBbits.RB7


#define LCD_CLR		0x01
#define LCD_CR		0x02
#define LCD_CON		0x0F

void lcd_clock() {
    LCD_E = 1;
    DLY();
    LCD_E = 0;
}

void write_lcd_data (char ch) {
	LCD_RS = 0;
	LCD_INPUT;
	LCD_RW = 1;
	do {
		LCD_E = 0;
		DLY();
		LCD_E = 1;
		DLY();
	} while (LCD_BF);
	LCD_E = 0;

	LCD_OUTPUT;

	LCD_RS = 1;
	LCD_RW = 0;
	LCD_OUT = ch;
	lcd_clock();
}

void write_lcd_cmd (char ch) {
	LCD_RS = 0;
	LCD_INPUT;
	LCD_RW = 1;
	do {
		LCD_E = 0;
		DLY();
		LCD_E = 1;
		DLY();
	} while (LCD_BF);
	LCD_E = 0;

	LCD_OUTPUT;

	LCD_RS = 0;
	LCD_RW = 0;
	LCD_OUT = ch;
	lcd_clock();
}

void init_lcd () {
	LCD_RS = 0;
	LCD_RW = 0;
	LCD_E = 0;
	LCD_OUT = 0x38;		// 8-bit operation, 2 lines 5x8 pt
	delay (15000);
	lcd_clock();
	delay (4100);
	lcd_clock();
	delay (100);
	lcd_clock();
	delay (50);
	lcd_clock();

	write_lcd_cmd(0x08);	// display off	
	write_lcd_cmd(0x01);	// clear display	
	write_lcd_cmd(0x06);	// increment, dont shift	
	write_lcd_cmd(0x0C);	// display on, no cursor
}

void clear_lcd() {
	write_lcd_cmd(0x01);	// clear display	
	write_lcd_cmd(0x80); 	// line 1	
}

void print_lcd_string (const char* str) {
	unsigned char count = 0;
	while(str[count] != '\0')
	{
	write_lcd_data(str[count]);
	count++;
	}
}

void print_lcd_word (unsigned int w) {
	K15WORD ww;
	ww.word = w;
	write_lcd_data(hex[ww.hi >> 4]);
	write_lcd_data(hex[ww.hi & 0x0F]);
	write_lcd_data(hex[ww.lo >> 4]);
	write_lcd_data(hex[ww.lo & 0x0F]);
}
