/**
 * Example on writing to the PIC16f628a's EEPROM memory.  This is 
 * tricky.  The exact sequence is:
 *
 * @ Select the correct address, enter data
 * @ Set the WREN flag, if it's not set already
 * @ Execute the correct 5-instruction sequence as given by Microchip
 *
 * That last part is hard to do in C since it needs to be done just 
 * right, so inline assembly is used.
 *
 * This example writes a string into EEPROM memory, starting at position 
 * 0x00.  The string is the name of this file and the date it was 
 * compiled.  To prove it's been written, you can read it back with your 
 * PIC programmer and see the difference.
 */
#include <pic18f252.h>
//#include "tsmtypes.h"

// Set the __CONFIG word:
//Uint16 at 0x2007  __CONFIG = CONFIG_WORD;

/** This EXACT SEQUENCE of instructions is needed, any deviation will
 *  cause the write to FAIL!  The PIC even counts the number of 
 *  instructions to check if you did this right!  This means we need 
 *  inline ASM, C isn't going to get it Just Right(tm).
 */
#define EEPROM_WRITE()	do {				\
	EECON2=0x00;					\
	__asm	MOVLW	0x55		__endasm; 	\
	__asm	MOVWF	_EECON2		__endasm; 	\
	__asm	MOVLW	0xaa		__endasm; 	\
	__asm	MOVWF	_EECON2		__endasm; 	\
	__asm	BSF	_EECON1,1;	__endasm;	\
	} while(0)


void eeprom_write_float(unsigned char address, float* value) {
	char* ch = (char*)value;
	char i;
	for (i=0; i<sizeof(float); i++) {
		EEADR=address++;	// EEPROM position same as string position.
		EEDATA= ch[i];		// Set the data to write.
		EECON1bits.WREN=1;			// Enable writes.
			EEPROM_WRITE();	// Magic Instruction Sequence Go!
			while(EECON1bits.WR);	// Wait for write to finish.
		EECON1bits.WREN=0;			// Disable writes.

	} 
}

void eeprom_read_float(unsigned char address, float* value) {
	char* ch = (char*)value;
	char i;
	for (i=0; i<sizeof(float); i++) {
		EEADR=address++;	// EEPROM position same as string position.
		EECON1bits.CFGS = 0;
		EECON1bits.EEPGD = 0;
		EECON1bits.RD = 1;
		ch[i] = EEDATA;
	} 
}

